#
# -*- coding: utf-8 -*-
#
# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""TEQ quantization."""

from typing import Any, List

import torch

from neural_compressor.torch.algorithms.base_algorithm import Quantizer
from neural_compressor.torch.utils import get_accelerator, get_model_device, is_transformers_imported, logger

from .modules import MulLinear, TEQLinearFakeQuant
from .utility import get_module, quant_tensor, set_module

if is_transformers_imported():
    import transformers

__all__ = ["TrainableEquivalentTransformation", "TEQuantizer"]


class TrainableEquivalentTransformation:
    """Weight-only quantization, Trainable Equivalent Transformation (TEQ)."""

    _PREPARE_ATTRS: List[str] = ["weight_config", "trained_alphas"]
    _PREPARE_ATTRS_PREFIX = "_prepare_"

    def __init__(self, model, weight_config={}, absorb_to_layer=None, folding=True, example_inputs=None):
        """Init the TrainableEquivalentTransformation object.

        Args:
            model (torch.nn.module): the model for quantization
            weight_config (dict, optional): contains all info required by RTN. Defaults to {}.
            absorb_to_layer (dict): The layer dict that scale can be absorbed. Default to None.
            folding(bool): Allow insert mul before linear when the scale cannot be absorbed by last layer.
                            Default to True.
            example_inputs: inputs for trace. Default to None.
        """
        self.model = model
        self.weight_config = weight_config
        self.folding = folding
        self.example_inputs = example_inputs
        self.device = self._get_device()
        self.trained_alphas = {}
        self.absorb_to_layer = absorb_to_layer
        self._post_initialized = False

    def _detect_absorb_to_layer(self, model, folding, example_inputs):
        # If user not provide the layers to absorb the quantization, detect layers automatically
        supported_layers = ["Linear"]
        detected_absorb_layers = {}
        # Detect the layers that can be absorbed automatically
        if folding:
            from neural_compressor.torch.algorithms.weight_only.utility import GraphTrace

            tg = GraphTrace()
            detected_absorb_layers, _ = tg.get_absorb_to_layer(model, example_inputs, supported_layers)
        else:  # pragma: no cover
            for name, module in model.named_modules():
                if module.__class__.__name__ in supported_layers:
                    detected_absorb_layers[name] = [name]
        logger.info("Detected **absorb layer**: **absorbed layers**")
        logger.info(detected_absorb_layers)
        return detected_absorb_layers

    def _post_init(self):
        self.dtype = self._get_dtype()
        self.model.to(self.device)
        self.model.eval()
        self._post_initialized = True

    def _get_device(self):
        """Get the model device.

        Returns:
            str: Model device.
        """
        device = get_accelerator().current_device_name()
        return device

    def _get_dtype(self):
        for _, p in self.model.named_parameters():
            return p.data.dtype

    def add_tuning_scale(self, sqrt_w_init=False):
        """Add tuning scales.

        Args:
            sqrt_w_init: use sqrt weight to init.
        """
        if not self.absorb_to_layer:
            self.absorb_to_layer = self._detect_absorb_to_layer(self.model, self.folding, self.example_inputs)
        if not self._post_initialized:
            self._post_init()
        # freeze model.
        for n, p in self.model.named_parameters():
            p.requires_grad = False

        for layer_norm in self.absorb_to_layer:
            layer_0_name = self.absorb_to_layer[layer_norm][0]

            module = get_module(self.model, layer_0_name)

            if sqrt_w_init:  # pragma: no cover
                weights = []
                for layer_name in self.absorb_to_layer[layer_norm]:
                    module = get_module(self.model, layer_name)
                    weights.append(module.weight)

                weights = torch.cat(weights, dim=0)
                max_value = torch.sqrt(torch.max(torch.abs(weights), dim=0).values)
                max_value[max_value == 0] = 1.0
                max_value = 1.0 / max_value

                alpha = torch.nn.Parameter(max_value)
                alpha = alpha.to(self.device)
            else:
                alpha = torch.nn.Parameter(torch.ones(module.weight.shape[1], device=self.device))

            self.trained_alphas[layer_norm] = alpha
            for layer_name in self.absorb_to_layer[layer_norm]:
                if not self.weight_config.get(layer_name):  # pragma: no cover
                    logger.info(f"layer {layer_name} not in weight config, skip.")
                    continue
                num_bits = self.weight_config[layer_name]["bits"]
                group_size = self.weight_config[layer_name]["group_size"]
                scheme = self.weight_config[layer_name]["scheme"]

                module = get_module(self.model, layer_name)
                wrapper_module = TEQLinearFakeQuant(
                    orig_layer=module, alpha=alpha, num_bits=num_bits, group_size=group_size, scheme=scheme
                )
                set_module(self.model, layer_name, wrapper_module)

        for layer_name, m in self.model.named_modules():
            if isinstance(m, torch.nn.Linear) and "orig_layer" not in n:
                if not self.weight_config.get(layer_name):  # pragma: no cover
                    logger.info(f"out of absorbed layer {layer_name} not in weight config, skip.")
                    continue
                num_bits = self.weight_config[layer_name]["bits"]
                group_size = self.weight_config[layer_name]["group_size"]
                scheme = self.weight_config[layer_name]["scheme"]

                alpha = torch.nn.Parameter(torch.ones(m.weight.shape[1], device=self.device))
                alpha.requires_grad_(False)
                wrapper_module = TEQLinearFakeQuant(
                    orig_layer=m, alpha=alpha, num_bits=num_bits, group_size=group_size, scheme=scheme
                )
                set_module(self.model, layer_name, wrapper_module)
        # Attach the weight config captured at prepare stage to the model
        self.model._weight_config = self.weight_config
        self.model._trained_alphas = self.trained_alphas

    @torch.no_grad()
    def _absorb_scales(self, layer, scale, layer_name=""):
        """Absorb the scale to the layer at output channel.

        Args:
            layer: the module.
            scale: the scale to be absorbed.
            layer_name: the layer name.
        """
        # for insert mul
        if not self.folding:  # pragma: no cover
            if isinstance(layer, MulLinear):
                set_module(self.model, layer_name, layer.linear)  ##recover
            else:
                new_module = MulLinear(layer, scale)
                set_module(self.model, layer_name, new_module)
            if not self.weight_config.get(layer_name):  # pragma: no cover
                logger.info(f"Absorb scale out of absorbed layer {layer_name} not in weight config, skip.")
                return
            self.weight_config[layer_name + ".linear"] = self.weight_config[layer_name]
            return

        if (
            isinstance(layer, torch.nn.BatchNorm2d)
            or isinstance(layer, torch.nn.GroupNorm)
            or isinstance(layer, torch.nn.InstanceNorm2d)
        ):
            if layer.affine:  # pragma: no cover
                layer.weight *= scale
                layer.bias *= scale
            else:  # pragma: no cover
                layer.affine = True
                weight = torch.ones(layer.num_features, device=self.device, dtype=self.dtype) * scale
                layer.weight = torch.nn.Parameter(weight, requires_grad=False)
                bias = torch.zeros(layer.num_features, device=self.device, dtype=self.dtype)
                layer.bias = torch.nn.Parameter(bias, requires_grad=False)
        elif isinstance(layer, torch.nn.LayerNorm):
            if layer.elementwise_affine:
                layer.weight *= scale
                layer.bias *= scale
            else:  # pragma: no cover
                layer.elementwise_affine = True
                weight = torch.ones(layer.num_features, device=self.device, dtype=self.dtype) * scale
                layer.weight = torch.nn.Parameter(torch.ones(weight, requires_grad=False))
                bias = torch.zeros(layer.num_features, device=self.device, dtype=self.dtype)
                layer.bias = torch.nn.Parameter(bias, requires_grad=False)

        elif isinstance(layer, torch.nn.Conv2d):  # pragma: no cover
            ## the order could not be changed
            if hasattr(layer, "bias") and (layer.bias is not None):
                layer.bias *= scale
            scale = scale.view(scale.shape[0], 1, 1, 1)
            layer.weight *= scale

        elif isinstance(layer, torch.nn.Linear):  # pragma: no cover
            if hasattr(layer, "bias") and (layer.bias is not None):
                layer.bias *= scale
            scale = scale.view(scale.shape[0], 1)
            layer.weight *= scale

        elif (
            layer.__class__.__name__ == "LlamaRMSNorm" or layer.__class__.__name__ == "T5LayerNorm"
        ):  # pragma: no cover
            layer.weight *= scale

        else:  # pragma: no cover
            logger.info(
                f"found unsupported layer {type(layer)}, try to multiply scale to "
                f"weight and bias directly, this may introduce accuracy issue, please have a check "
            )
            if hasattr(layer, "weight") and layer.weight is not None:
                layer.weight *= scale
            if hasattr(layer, "bias") and layer.bias is not None:
                layer.bias *= scale

    @torch.no_grad()
    def _scale_layer_weight(self, layer, scale):  ##input channel
        """Scale the layer weights at input channel, depthwise conv output channel.

        Args:
            layer: the layer.
            scale: the scale to be multiplied.
        """
        if layer.__class__.__name__ == "MulLinear":
            layer = layer.linear

        if layer.__class__.__name__ == "TEQLinearFakeQuant":
            layer = layer.orig_layer

        scale = scale.view(1, scale.shape[0])
        layer.weight = torch.nn.Parameter(layer.weight * scale)
        return scale

    @torch.no_grad()
    def transform(self):
        """Apply alpha/scale."""
        if not self._post_initialized:  # pragma: no cover
            self._post_init()
        for ln_name, layer_names in self.absorb_to_layer.items():
            module = get_module(self.model, ln_name)
            scale = self.trained_alphas[ln_name]
            scale = torch.clip(scale, 1e-5)
            input_scale = 1.0 / scale
            if hasattr(module, "orig_layer"):
                module = module.orig_layer

            self._absorb_scales(module, input_scale, layer_name=ln_name)
            weight_scale = scale
            for layer_name in layer_names:
                layer_module = get_module(self.model, layer_name)
                self._scale_layer_weight(layer_module, weight_scale)

        # for Folding = True
        for n, m in self.model.named_modules():
            if isinstance(m, TEQLinearFakeQuant):
                set_module(self.model, n, m.orig_layer)

    @torch.no_grad()
    def quantize(self, **kwargs):
        """quantization."""
        use_optimum_format = kwargs.get("use_optimum_format", True)
        device = get_accelerator().current_device_name()
        model_device = get_model_device(self.model)  # return model on the same device
        model = self.model
        for name, m in model.named_modules():
            if self.weight_config.get(name) is None:  # pragma: no cover
                logger.info(f"quantize layer {name} not in weight config, skip.")
                continue
            num_bits = self.weight_config[name]["bits"]
            group_size = self.weight_config[name]["group_size"]
            scheme = self.weight_config[name]["scheme"]
            group_dim = self.weight_config[name].get("group_dim", 1)
            # for only group_dim is 0 or only `transformers.Conv1D`, we need transpose weight.
            if is_transformers_imported():
                transpose = (group_dim == 0) ^ (isinstance(m, transformers.Conv1D))
            else:  # pragma: no cover
                transpose = group_dim == 0
            if transpose:  # pragma: no cover
                weight = m.weight.detach().T.contiguous()
            else:
                weight = m.weight.detach()
            if isinstance(m, torch.nn.Linear):  # pragma: no cover
                int_weight, scale, zp = quant_tensor(
                    weight.data,
                    num_bits=num_bits,
                    group_size=group_size,
                    scheme=scheme,
                    return_int=True,
                )
                int_weight = int_weight.t_().contiguous() if transpose else int_weight
                scale = scale.t_().contiguous() if transpose else scale
                zp = zp.t_().contiguous() if transpose and zp is not None else zp
                if isinstance(m, torch.nn.Linear):
                    in_features = m.in_features
                    out_features = m.out_features
                elif is_transformers_imported() and isinstance(m, transformers.Conv1D):
                    in_features = m.weight.shape[0]
                    out_features = m.weight.shape[1]
                    int_weight = int_weight.t_().contiguous()
                    scale = scale.t_().contiguous()
                    zp = zp.t_().contiguous() if zp is not None else zp
                from .modules import INCWeightOnlyLinear

                new_module = INCWeightOnlyLinear(
                    in_features,
                    out_features,
                    bits=num_bits,
                    group_size=group_size,
                    zp=zp is not None,
                    bias=m.bias is not None,
                    use_optimum_format=use_optimum_format,
                    device=device,
                )
                new_module.pack(int_weight, scale, zp, m.bias)
                if name == "":
                    return new_module
                else:
                    set_module(model, name, new_module)
                # Move modules back to the model device layer-by-layer
                m.to(model_device)
                new_module.to(model_device)
        self.model = model

    def save(self, save_scale_file="", save_state_dict_file=""):
        """Save alpha/scale or model weight.

        Args:
            save_scale_file: path to save alpha/scale with torch.save.
            save_state_dict_file: path to save model state_dict.
        """
        if save_scale_file:  # pragma: no cover
            torch.save(self.trained_alphas, save_scale_file)

        if save_state_dict_file:  # pragma: no cover
            torch.save(self.model.state_dict(), save_state_dict_file)


class TEQuantizer(Quantizer):
    """TEQ Quantizer."""

    def __init__(self, quant_config, folding, example_inputs, absorb_to_layer=None):
        """Init the TEQuantizer object."""
        super().__init__(quant_config=quant_config)
        self.folding = folding
        self.absorb_to_layer = absorb_to_layer
        self.example_inputs = example_inputs
        self._quantizer = TrainableEquivalentTransformation(
            model=None,
            weight_config=quant_config,
            absorb_to_layer=absorb_to_layer,
            folding=folding,
            example_inputs=example_inputs,
        )

    def prepare(self, model, *args, **kwargs):
        """Prepares a given model for quantization.

        Args:
            model: A float model to be quantized.

        Returns:
            A prepared model.
        """
        float_model = model
        assert isinstance(model, torch.nn.Module), "only support torch module"
        self._quantizer.model = float_model
        logger.info("TEQ quantizing start.")
        self._quantizer.add_tuning_scale()
        for attr in self._quantizer._PREPARE_ATTRS:
            setattr(float_model, self._quantizer._PREPARE_ATTRS_PREFIX + attr, getattr(self._quantizer, attr))
        return float_model

    def convert(self, model, *args: Any, **kwargs: Any):
        """Convert the prepared model to a quantized model.

        Args:
            model (torch.nn.Module): the prepared model

        Returns:
            The quantized model.
        """
        for attr in self._quantizer._PREPARE_ATTRS:
            setattr(self._quantizer, attr, getattr(model, self._quantizer._PREPARE_ATTRS_PREFIX + attr, None))
        self._quantizer.model = model
        self._quantizer.transform()
        self._quantizer.quantize(**kwargs)
        logger.info("TEQ quantizing done.")
        return self._quantizer.model



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import time
import warnings
from typing import List, Optional, Tuple, Union

import torch
import torch.distributed as dist
from transformers.generation.logits_process import LogitsProcessorList
from transformers.generation.stopping_criteria import StoppingCriteriaList, validate_stopping_criteria
from transformers.generation.streamers import BaseStreamer
from transformers.utils import ModelOutput


class GreedySearchDecoderOnlyOutput(ModelOutput):
    sequences: torch.LongTensor = None
    scores: Optional[Tuple[torch.FloatTensor]] = None
    attentions: Optional[Tuple[Tuple[torch.FloatTensor]]] = None
    hidden_states: Optional[Tuple[Tuple[torch.FloatTensor]]] = None
    past_key_values: Optional[Tuple[Tuple[Tuple[torch.FloatTensor]]]] = None


class GreedySearchEncoderDecoderOutput(ModelOutput):
    sequences: torch.LongTensor = None
    scores: Optional[Tuple[torch.FloatTensor]] = None
    encoder_attentions: Optional[Tuple[torch.FloatTensor]] = None
    encoder_hidden_states: Optional[Tuple[torch.FloatTensor]] = None
    decoder_attentions: Optional[Tuple[Tuple[torch.FloatTensor]]] = None
    cross_attentions: Optional[Tuple[Tuple[torch.FloatTensor]]] = None
    decoder_hidden_states: Optional[Tuple[Tuple[torch.FloatTensor]]] = None
    past_key_values: Optional[Tuple[Tuple[Tuple[torch.FloatTensor]]]] = None


GreedySearchOutput = Union[GreedySearchEncoderDecoderOutput, GreedySearchDecoderOnlyOutput]


def _greedy_search(
    self,
    input_ids: torch.LongTensor,
    logits_processor: Optional[LogitsProcessorList] = None,
    stopping_criteria: Optional[StoppingCriteriaList] = None,
    max_length: Optional[int] = None,
    pad_token_id: Optional[int] = None,
    eos_token_id: Optional[Union[int, List[int]]] = None,
    output_attentions: Optional[bool] = None,
    output_hidden_states: Optional[bool] = None,
    output_scores: Optional[bool] = None,
    return_dict_in_generate: Optional[bool] = None,
    synced_gpus: bool = False,
    streamer: Optional["BaseStreamer"] = None,
    **model_kwargs,
) -> Union[GreedySearchOutput, torch.LongTensor]:
    r"""Generates sequences of token ids for models with a language modeling head using **greedy decoding** and can be
    used for text-decoder, text-to-text, speech-to-text, and vision-to-text models.

    <Tip warning={true}>
    In most cases, you do not need to call [`~generation.GenerationMixin.greedy_search`] directly. Use generate()
    instead. For an overview of generation strategies and code examples, check the [following
    guide](../generation_strategies).
    </Tip>
    Parameters:
        input_ids (`torch.LongTensor` of shape `(batch_size, sequence_length)`):
            The sequence used as a prompt for the generation.
        logits_processor (`LogitsProcessorList`, *optional*):
            An instance of [`LogitsProcessorList`]. List of instances of class derived from [`LogitsProcessor`]
            used to modify the prediction scores of the language modeling head applied at each generation step.
        stopping_criteria (`StoppingCriteriaList`, *optional*):
            An instance of [`StoppingCriteriaList`]. List of instances of class derived from [`StoppingCriteria`]
            used to tell if the generation loop should stop.
        max_length (`int`, *optional*, defaults to 20):
            **DEPRECATED**. Use `logits_processor` or `stopping_criteria` directly to cap the number of generated
            tokens. The maximum length of the sequence to be generated.
        pad_token_id (`int`, *optional*):
            The id of the *padding* token.
        eos_token_id (`Union[int, List[int]]`, *optional*):
            The id of the *end-of-sequence* token. Optionally, use a list to set multiple *end-of-sequence* tokens.
        output_attentions (`bool`, *optional*, defaults to `False`):
            Whether or not to return the attentions tensors of all attention layers. See `attentions` under
            returned tensors for more details.
        output_hidden_states (`bool`, *optional*, defaults to `False`):
            Whether or not to return the hidden states of all layers. See `hidden_states` under returned tensors
            for more details.
        output_scores (`bool`, *optional*, defaults to `False`):
            Whether or not to return the prediction scores. See `scores` under returned tensors for more details.
        return_dict_in_generate (`bool`, *optional*, defaults to `False`):
            Whether or not to return a [`~utils.ModelOutput`] instead of a plain tuple.
        synced_gpus (`bool`, *optional*, defaults to `False`):
            Whether to continue running the while loop until max_length (needed for ZeRO stage 3)
        streamer (`BaseStreamer`, *optional*):
            Streamer object that will be used to stream the generated sequences. Generated tokens are passed
            through `streamer.put(token_ids)` and the streamer is responsible for any further processing.
        model_kwargs:
            Additional model specific keyword arguments will be forwarded to the `forward` function of the model.
            If model is an encoder-decoder model the kwargs should include `encoder_outputs`.
    Return:
        [`GreedySearchDecoderOnlyOutput`], [`GreedySearchEncoderDecoderOutput`] or
        `torch.LongTensor`: A `torch.LongTensor` containing the generated tokens (default behaviour) or a
        [`GreedySearchDecoderOnlyOutput`] if `model.config.is_encoder_decoder=False` and
        `return_dict_in_generate=True` or a [`GreedySearchEncoderDecoderOutput`] if
        `model.config.is_encoder_decoder=True`.
    Examples:
    ```python
    >>> from transformers import (
    ...     AutoTokenizer,
    ...     AutoModelForCausalLM,
    ...     LogitsProcessorList,
    ...     MinLengthLogitsProcessor,
    ...     StoppingCriteriaList,
    ...     MaxLengthCriteria,
    ... )
    >>> tokenizer = AutoTokenizer.from_pretrained("gpt2")
    >>> model = AutoModelForCausalLM.from_pretrained("gpt2")
    >>> # set pad_token_id to eos_token_id because GPT2 does not have a PAD token
    >>> model.generation_config.pad_token_id = model.generation_config.eos_token_id
    >>> input_prompt = "It might be possible to"
    >>> input_ids = tokenizer(input_prompt, return_tensors="pt").input_ids
    >>> # instantiate logits processors
    >>> logits_processor = LogitsProcessorList(
    ...     [
    ...         MinLengthLogitsProcessor(10, eos_token_id=model.generation_config.eos_token_id),
    ...     ]
    ... )
    >>> stopping_criteria = StoppingCriteriaList([MaxLengthCriteria(max_length=20)])
    >>> outputs = model.greedy_search(
    ...     input_ids, logits_processor=logits_processor, stopping_criteria=stopping_criteria
    ... )
    >>> tokenizer.batch_decode(outputs, skip_special_tokens=True)
    ["It might be possible to get a better understanding of the nature of the problem, but it's not"]
    ```
    """
    token_latency = (self.config.token_latency if hasattr(self.config, "token_latency") else False) or (
        self.token_latency if hasattr(self, "token_latency") else False
    )

    latency_list = []
    # init values
    logits_processor = logits_processor if logits_processor is not None else LogitsProcessorList()
    stopping_criteria = stopping_criteria if stopping_criteria is not None else StoppingCriteriaList()
    if max_length is not None:
        warnings.warn(
            "`max_length` is deprecated in this function, use"
            " `stopping_criteria=StoppingCriteriaList([MaxLengthCriteria(max_length=max_length)])` instead.",
            UserWarning,
        )
        stopping_criteria = validate_stopping_criteria(stopping_criteria, max_length)
    pad_token_id = pad_token_id if pad_token_id is not None else self.generation_config.pad_token_id
    eos_token_id = eos_token_id if eos_token_id is not None else self.generation_config.eos_token_id
    if isinstance(eos_token_id, int):
        eos_token_id = [eos_token_id]
    eos_token_id_tensor = torch.tensor(eos_token_id).to(input_ids.device) if eos_token_id is not None else None
    output_scores = output_scores if output_scores is not None else self.generation_config.output_scores
    output_attentions = output_attentions if output_attentions is not None else self.generation_config.output_attentions
    output_hidden_states = (
        output_hidden_states if output_hidden_states is not None else self.generation_config.output_hidden_states
    )
    return_dict_in_generate = (
        return_dict_in_generate
        if return_dict_in_generate is not None
        else self.generation_config.return_dict_in_generate
    )

    # init attention / hidden states / scores tuples
    scores = () if (return_dict_in_generate and output_scores) else None
    decoder_attentions = () if (return_dict_in_generate and output_attentions) else None
    cross_attentions = () if (return_dict_in_generate and output_attentions) else None
    decoder_hidden_states = () if (return_dict_in_generate and output_hidden_states) else None

    # if model is an encoder-decoder, retrieve encoder attention weights and hidden states
    if return_dict_in_generate and self.config.is_encoder_decoder:
        encoder_attentions = model_kwargs["encoder_outputs"].get("attentions") if output_attentions else None
        encoder_hidden_states = model_kwargs["encoder_outputs"].get("hidden_states") if output_hidden_states else None
    # keep track of which sequences are already finished
    unfinished_sequences = torch.ones(input_ids.shape[0], dtype=torch.long, device=input_ids.device)
    this_peer_finished = False  # used by synced_gpus only
    while True:
        tic = time.time()
        if synced_gpus:
            # Under synced_gpus the `forward` call must continue until all gpus complete their sequence.
            # The following logic allows an early break if all peers finished generating their sequence
            this_peer_finished_flag = torch.tensor(0.0 if this_peer_finished else 1.0).to(input_ids.device)
            # send 0.0 if we finished, 1.0 otherwise
            dist.all_reduce(this_peer_finished_flag, op=dist.ReduceOp.SUM)
            # did all peers finish? the reduced sum will be 0.0 then
            if this_peer_finished_flag.item() == 0.0:
                break

        # prepare model inputs
        model_inputs = self.prepare_inputs_for_generation(input_ids, **model_kwargs)
        if (
            re.search("GPTJ", self.config.architectures[0])
            or re.search("llama", self.config.architectures[0], re.IGNORECASE)
            or re.search("gptneox", self.config.architectures[0], re.IGNORECASE)
            or re.search("OPT", self.config.architectures[0], re.IGNORECASE)
            or re.search("falcon", self.config.architectures[0], re.IGNORECASE)
            or re.search("rw", self.config.architectures[0], re.IGNORECASE)
        ):
            first_token = False
            input_bs = input_ids.size()[0]
            if model_inputs["past_key_values"] is None:
                first_token = True
            if first_token and hasattr(self, "trace_graph"):
                if re.search("GPTJ", self.config.architectures[0]):
                    beam_idx_tmp = torch.zeros(
                        (2048, int(input_bs)), dtype=torch.long, device=input_ids.device
                    ).contiguous()
                    model_inputs["past_key_values"] = tuple(
                        [
                            (
                                torch.zeros(1, 0, 0, 1, dtype=torch.long, device=input_ids.device).contiguous(),
                                torch.zeros([1, 1, 1, 1], device=input_ids.device).contiguous(),
                                torch.zeros([1, 1, 1, 1], device=input_ids.device).contiguous(),
                                beam_idx_tmp,
                            )
                            for i in range(self.config.n_layer)
                        ]
                    )
                elif re.search("llama", self.config.architectures[0], re.IGNORECASE):
                    beam_idx_tmp = torch.zeros(
                        (2048, int(input_bs)), dtype=torch.long, device=input_ids.device
                    ).contiguous()
                    model_inputs["past_key_values"] = tuple(
                        [
                            (
                                torch.zeros(1, 0, 0, 1, dtype=torch.long, device=input_ids.device).contiguous(),
                                torch.zeros([1, 1, 1, 1], device=input_ids.device).contiguous(),
                                torch.zeros([1, 1, 1, 1], device=input_ids.device).contiguous(),
                                beam_idx_tmp,
                            )
                            for i in range(self.config.num_hidden_layers)
                        ]
                    )
                elif re.search("gptneox", self.config.architectures[0], re.IGNORECASE):
                    beam_idx_tmp = torch.zeros(
                        (2048, int(input_bs)), dtype=torch.long, device=input_ids.device
                    ).contiguous()
                    model_inputs["past_key_values"] = tuple(
                        [
                            (
                                torch.zeros(1, 0, 0, 1, dtype=torch.long, device=input_ids.device).contiguous(),
                                torch.zeros([1, 1, 1, 1], device=input_ids.device).contiguous(),
                                torch.zeros([1, 1, 1, 1], device=input_ids.device).contiguous(),
                                beam_idx_tmp,
                            )
                            for i in range(self.config.num_hidden_layers)
                        ]
                    )
                elif re.search("OPT", self.config.architectures[0], re.IGNORECASE):
                    beam_idx_tmp = torch.zeros(
                        (2048, int(input_bs)), dtype=torch.long, device=input_ids.device
                    ).contiguous()
                    model_inputs["past_key_values"] = tuple(
                        [
                            (
                                torch.zeros(1, 0, 0, 1, dtype=torch.long, device=input_ids.device).contiguous(),
                                torch.zeros([1, 1, 1, 1], device=input_ids.device).contiguous(),
                                torch.zeros([1, 1, 1, 1], device=input_ids.device).contiguous(),
                                beam_idx_tmp,
                            )
                            for i in range(self.config.num_hidden_layers)
                        ]
                    )
                elif re.search("falcon", self.config.architectures[0], re.IGNORECASE) or re.search(
                    "rw", self.config.architectures[0], re.IGNORECASE
                ):
                    beam_idx_tmp = torch.zeros(
                        (2048, int(input_bs)), dtype=torch.long, device=input_ids.device
                    ).contiguous()
                    model_inputs["past_key_values"] = tuple(
                        [
                            (
                                torch.zeros(1, 0, 0, 1, dtype=torch.long, device=input_ids.device).contiguous(),
                                torch.zeros([1, 1, 1, 1], device=input_ids.device).contiguous(),
                                torch.zeros([1, 1, 1, 1], device=input_ids.device).contiguous(),
                                beam_idx_tmp,
                            )
                            for i in range(self.config.num_hidden_layers)
                        ]
                    )
            if hasattr(self, "trace_graph"):
                model_inputs.pop("use_cache", None)
                model_inputs.pop("token_type_ids", None)
                outputs = self.trace_graph(**model_inputs)
                if synced_gpus and this_peer_finished:
                    continue  # don't waste resources running the code we don't need
                next_token_logits = outputs[0][:, -1, :]
            else:
                outputs = self(
                    **model_inputs,
                    return_dict=True,
                    output_attentions=output_attentions,
                    output_hidden_states=output_hidden_states,
                )
                if synced_gpus and this_peer_finished:
                    continue  # don't waste resources running the code we don't need
                next_token_logits = outputs.logits[:, -1, :]
        else:
            outputs = self(
                **model_inputs,
                return_dict=True,
                output_attentions=output_attentions,
                output_hidden_states=output_hidden_states,
            )
            if synced_gpus and this_peer_finished:
                continue  # don't waste resources running the code we don't need
            next_token_logits = outputs.logits[:, -1, :]

        # pre-process distribution
        next_tokens_scores = logits_processor(input_ids, next_token_logits)

        # Store scores, attentions and hidden_states when required
        if return_dict_in_generate:
            if output_scores:
                scores += (next_tokens_scores,)
            if output_attentions:
                decoder_attentions += (
                    (outputs.decoder_attentions,) if self.config.is_encoder_decoder else (outputs.attentions,)
                )
                if self.config.is_encoder_decoder:
                    cross_attentions += (outputs.cross_attentions,)

            if output_hidden_states:
                decoder_hidden_states += (
                    (outputs.decoder_hidden_states,) if self.config.is_encoder_decoder else (outputs.hidden_states,)
                )

        # argmax
        next_tokens = torch.argmax(next_tokens_scores, dim=-1)

        # finished sentences should have their next token be a padding token
        if eos_token_id is not None:
            if pad_token_id is None:
                raise ValueError("If `eos_token_id` is defined, make sure that `pad_token_id` is defined.")
            next_tokens = next_tokens * unfinished_sequences + pad_token_id * (1 - unfinished_sequences)
        # update generated ids, model inputs, and length for next step
        input_ids = torch.cat([input_ids, next_tokens[:, None]], dim=-1)
        if streamer is not None:
            streamer.put(next_tokens.cpu())
        model_kwargs = self._update_model_kwargs_for_generation(
            outputs, model_kwargs, is_encoder_decoder=self.config.is_encoder_decoder
        )

        # if eos_token was found in one sentence, set sentence to finished
        if eos_token_id_tensor is not None:
            unfinished_sequences = unfinished_sequences.mul(
                next_tokens.tile(eos_token_id_tensor.shape[0], 1).ne(eos_token_id_tensor.unsqueeze(1)).prod(dim=0)
            )
            # stop when each sentence is finished
            if unfinished_sequences.max() == 0:
                this_peer_finished = True
        # stop if we exceed the maximum length
        if token_latency:
            if input_ids.is_xpu:
                torch.xpu.synchronize()
            latency_list.append(time.time() - tic)
        if stopping_criteria(input_ids, scores):
            this_peer_finished = True
        if this_peer_finished and not synced_gpus:
            break
    if streamer is not None:
        streamer.end()

    if return_dict_in_generate:
        if self.config.is_encoder_decoder:
            output_result = GreedySearchEncoderDecoderOutput(
                sequences=input_ids,
                scores=scores,
                encoder_attentions=encoder_attentions,
                encoder_hidden_states=encoder_hidden_states,
                decoder_attentions=decoder_attentions,
                cross_attentions=cross_attentions,
                decoder_hidden_states=decoder_hidden_states,
                past_key_values=model_kwargs.get("past_key_values"),
            )
        else:
            output_result = GreedySearchDecoderOnlyOutput(
                sequences=input_ids,
                scores=scores,
                attentions=decoder_attentions,
                hidden_states=decoder_hidden_states,
                past_key_values=model_kwargs.get("past_key_values"),
            )
    else:
        output_result = input_ids

    if token_latency:
        return (output_result, latency_list)
    else:
        return output_result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Load one specify tensor from a bin file."""

import io
import os
import warnings
from typing import IO, Any, BinaryIO, Callable, Dict, Optional, Union

from packaging.version import Version
from torch.serialization import (
    StorageType,
    _get_restore_location,
    _is_torchscript_zip,
    _is_zipfile,
    _maybe_decode_ascii,
    _open_file_like,
    _open_zipfile_reader,
)

from neural_compressor.torch.algorithms.layer_wise import modified_pickle as pickle
from neural_compressor.torch.utils import logger

from .utils import torch

torch_version = torch.__version__.split("+")[0]
version = Version(torch_version)

FILE_LIKE = Union[str, os.PathLike, BinaryIO, IO[bytes]]
MAP_LOCATION = Optional[Union[Callable[[torch.Tensor, str], torch.Tensor], torch.device, str, Dict[str, str]]]

if version.release < Version("1.13.0").release:
    UntypedStorage = torch._UntypedStorage
else:
    UntypedStorage = torch.UntypedStorage


def _load(zip_file, tensor_name, prefix, map_location, pickle_module, pickle_file="data.pkl", **pickle_load_args):
    restore_location = _get_restore_location(map_location)

    loaded_storages = {}

    def load_tensor(dtype, numel, key, location):
        name = f"data/{key}"

        if version.release < Version("1.13.0").release:
            storage = zip_file.get_storage_from_record(name, numel, torch._UntypedStorage).storage()._untyped()
            typed_storage = torch.storage._TypedStorage(wrap_storage=restore_location(storage, location), dtype=dtype)
            loaded_storages[key] = typed_storage
        elif version.release < Version("2.0.0").release:  # pragma: no cover
            storage = zip_file.get_storage_from_record(name, numel, UntypedStorage).storage().untyped()
            typed_storage = torch.storage.TypedStorage(wrap_storage=restore_location(storage, location), dtype=dtype)
            loaded_storages[key] = typed_storage
        else:
            storage = zip_file.get_storage_from_record(name, numel, UntypedStorage)._typed_storage()._untyped_storage
            typed_storage = torch.storage.TypedStorage(
                wrap_storage=restore_location(storage, location), dtype=dtype, _internal=True
            )

            if typed_storage._data_ptr() != 0:
                loaded_storages[key] = typed_storage

        return typed_storage

    load_module_mapping: Dict[str, str] = {"torch.tensor": "torch._tensor"}

    # Need to subclass Unpickler instead of directly monkey-patching the find_class method
    # because it's marked readonly in pickle.
    # The type: ignore is because mypy can't statically determine the type of this class.
    class UnpicklerWrapper(pickle_module.Unpickler):  # type: ignore[name-defined]
        def find_class(self, mod_name, name):
            if type(name) is str and "Storage" in name:
                try:
                    return StorageType(name)
                except KeyError:  # pragma: no cover
                    pass
            mod_name = load_module_mapping.get(mod_name, mod_name)
            return super().find_class(mod_name, name)

        def persistent_load(self, saved_id):
            assert isinstance(saved_id, tuple)
            typename = _maybe_decode_ascii(saved_id[0])
            data = saved_id[1:]

            assert (
                typename == "storage"
            ), f"Unknown typename for persistent_load, expected 'storage' but got '{typename}'"
            storage_type, key, location, numel = data

            if storage_type is UntypedStorage:  # pragma: no cover
                dtype = torch.uint8
            else:
                dtype = storage_type.dtype

            if key in loaded_storages:
                typed_storage = loaded_storages[key]
            else:
                name_list = [self.tensor_name]
                if prefix:
                    no_prefix_name = self.tensor_name.split(".")
                    if prefix in no_prefix_name:
                        no_prefix_name.remove(prefix)
                    no_prefix_name = ".".join(no_prefix_name)
                    name_list.append(no_prefix_name)
                if self.tensor_name and self.metastack[-1][-2] not in name_list:
                    # typed_storage = None
                    # loaded_storages[key] = typed_storage
                    # nbytes = numel * torch._utils._element_size(dtype)
                    # typed_storage = load_tensor(dtype, nbytes, key, _maybe_decode_ascii(location))
                    typed_storage = None
                else:
                    nbytes = numel * torch._utils._element_size(dtype)
                    typed_storage = load_tensor(dtype, nbytes, key, _maybe_decode_ascii(location))

            return typed_storage

    # Load the data (which may in turn use `persistent_load` to load tensors)
    data_file = io.BytesIO(zip_file.get_record(pickle_file))

    unpickler = UnpicklerWrapper(data_file, **pickle_load_args)
    # unpickler.persistent_load = persistent_load
    result = unpickler.load(tensor_name)

    torch._utils._validate_loaded_sparse_tensors()
    return result


def load(
    f: FILE_LIKE,
    tensor_name: str = None,
    prefix: str = None,
    map_location: MAP_LOCATION = None,
    pickle_module: Any = None,
    *,
    weights_only: bool = False,
    **pickle_load_args: Any,
) -> Any:
    # Reference: https://github.com/pytorch/pytorch/issues/54354
    # The first line of this docstring overrides the one Sphinx generates for the
    # documentation. We need it so that Sphinx doesn't leak `pickle`s path from
    # the build environment (e.g. `<module 'pickle' from '/leaked/path').
    """Load(f, map_location=None, pickle_module=pickle, *, weights_only=False, **pickle_load_args).

    Loads an object saved with :func:`torch.save` from a file.

    :func:`torch.load` uses Python's unpickling facilities but treats storages,
    which underlie tensors, specially. They are first deserialized on the
    CPU and are then moved to the device they were saved from. If this fails
    (e.g. because the run time system doesn't have certain devices), an exception
    is raised. However, storages can be dynamically remapped to an alternative
    set of devices using the :attr:`map_location` argument.

    If :attr:`map_location` is a callable, it will be called once for each serialized
    storage with two arguments: storage and location. The storage argument
    will be the initial deserialization of the storage, residing on the CPU.
    Each serialized storage has a location tag associated with it which
    identifies the device it was saved from, and this tag is the second
    argument passed to :attr:`map_location`. The builtin location tags are ``'cpu'``
    for CPU tensors and ``'cuda:device_id'`` (e.g. ``'cuda:2'``) for CUDA tensors.
    :attr:`map_location` should return either ``None`` or a storage. If
    :attr:`map_location` returns a storage, it will be used as the final deserialized
    object, already moved to the right device. Otherwise, :func:`torch.load` will
    fall back to the default behavior, as if :attr:`map_location` wasn't specified.

    If :attr:`map_location` is a :class:`torch.device` object or a string containing
    a device tag, it indicates the location where all tensors should be loaded.

    Otherwise, if :attr:`map_location` is a dict, it will be used to remap location tags
    appearing in the file (keys), to ones that specify where to put the
    storages (values).

    User extensions can register their own location tags and tagging and
    deserialization methods using :func:`torch.serialization.register_package`.

    Args:
        f: a file-like object (has to implement :meth:`read`, :meth:`readline`, :meth:`tell`, and :meth:`seek`),
            or a string or os.PathLike object containing a file name
        map_location: a function, :class:`torch.device`, string or a dict specifying how to remap storage
            locations
        pickle_module: module used for unpickling metadata and objects (has to
            match the :attr:`pickle_module` used to serialize file)
        weights_only: Indicates whether unpickler should be restricted to
            loading only tensors, primitive types and dictionaries
        pickle_load_args: (Python 3 only) optional keyword arguments passed over to
            :func:`pickle_module.load` and :func:`pickle_module.Unpickler`, e.g.,
            :attr:`errors=...`.
        prefix (str): the module prefix name.
        tensor_name (str): the tensor name.

    .. warning::
        :func:`torch.load()` unless `weights_only` parameter is set to `True`,
        uses ``pickle`` module implicitly, which is known to be insecure.
        It is possible to construct malicious pickle data which will execute arbitrary code
        during unpickling. Never load data that could have come from an untrusted
        source in an unsafe mode, or that could have been tampered with. **Only load data you trust**.

    .. note::
        When you call :func:`torch.load()` on a file which contains GPU tensors, those tensors
        will be loaded to GPU by default. You can call ``torch.load(.., map_location='cpu')``
        and then :meth:`load_state_dict` to avoid GPU RAM surge when loading a model checkpoint.

    .. note::
        By default, we decode byte strings as ``utf-8``.  This is to avoid a common error
        case ``UnicodeDecodeError: 'ascii' codec can't decode byte 0x...``
        when loading files saved by Python 2 in Python 3.  If this default
        is incorrect, you may use an extra :attr:`encoding` keyword argument to specify how
        these objects should be loaded, e.g., :attr:`encoding='latin1'` decodes them
        to strings using ``latin1`` encoding, and :attr:`encoding='bytes'` keeps them
        as byte arrays which can be decoded later with ``byte_array.decode(...)``.

    Example:
        >>> # xdoctest: +SKIP("undefined filepaths")
        >>> torch.load('tensors.pt')
        # Load all tensors onto the CPU
        >>> torch.load('tensors.pt', map_location=torch.device('cpu'))
        # Load all tensors onto the CPU, using a function
        >>> torch.load('tensors.pt', map_location=lambda storage, loc: storage)
        # Load all tensors onto GPU 1
        >>> torch.load('tensors.pt', map_location=lambda storage, loc: storage.cuda(1))
        # Map tensors from GPU 1 to GPU 0
        >>> torch.load('tensors.pt', map_location={'cuda:1': 'cuda:0'})
        # Load tensor from io.BytesIO object
        >>> with open('tensor.pt', 'rb') as f:
        ...     buffer = io.BytesIO(f.read())
        >>> torch.load(buffer)
        # Load a module with 'ascii' encoding for unpickling
        >>> torch.load('module.pt', encoding='ascii')
    """
    torch._C._log_api_usage_once("torch.load")

    # Security warning for layer-wise quantization checkpoint loading
    logger.warning_once(
        "Layer-wise quantization is loading a model checkpoint. "
        "Please ensure you trust the source of the checkpoint. "
        "Loading checkpoints from untrusted sources may execute arbitrary code.",
    )

    # Add ability to force safe only weight loads via environment variable
    if os.getenv("TORCH_FORCE_WEIGHTS_ONLY_LOAD", "0").lower() in ["1", "y", "yes", "true"]:  # pragma: no cover
        weights_only = True

    if weights_only:  # pragma: no cover
        if pickle_module is not None:
            raise RuntimeError("Can not safely load weights when explicit pickle_module is specified")
    else:
        if pickle_module is None:
            pickle_module = pickle

    if "encoding" not in pickle_load_args.keys():
        pickle_load_args["encoding"] = "utf-8"

    with _open_file_like(f, "rb") as opened_file:
        if _is_zipfile(opened_file):
            # The zipfile reader is going to advance the current file position.
            # If we want to actually tail call to torch.jit.load, we need to
            # reset back to the original position.
            orig_position = opened_file.tell()
            with _open_zipfile_reader(opened_file) as opened_zipfile:
                if _is_torchscript_zip(opened_zipfile):  # pragma: no cover
                    warnings.warn(
                        "'torch.load' received a zip file that looks like a TorchScript archive"
                        " dispatching to 'torch.jit.load' (call 'torch.jit.load' directly to"
                        " silence this warning)",
                        UserWarning,
                    )
                    opened_file.seek(orig_position)
                    return torch.jit.load(opened_file, map_location=map_location)
                return _load(opened_zipfile, tensor_name, prefix, map_location, pickle_module, **pickle_load_args)



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import torch
import torch.nn as nn
import numpy as np
import os

from .scale_methods import ops_quantizer
from .._quant_common.quant_config import QuantMode
from .._quant_common.helper_modules import PatchedUnmeasuredModule
# TODO [SW-217813]: support dynamic quantization in all ops and remove is_supported_dynamic_op
from .._quant_common.quant_config import get_hqt_config, set_hqt_config, is_supported_dynamic_op
from ..utils.logger import logger
from .common import convert_scales_to_tensors_dict, save_scales, load_scales
from .patching_common import generate_model_info, mod_default_dict, parent_child_mod_dict
from .measure import load_measurements
from .scale import load_layer_scales
from neural_compressor.torch.utils.auto_accelerator import auto_detect_accelerator
from neural_compressor.torch.algorithms.fp8_quant._core.common import dequant_original_fp8_weight_if_needed
from .scale_methods.scale_method_factory import ScaleValueType
from .scale_methods.scale_method_config import ScaleMethodConfig, find_node_scale_method_config, CfgStr, dump_scale_method_config_by_mod_map

cur_accelerator = auto_detect_accelerator()


def patch_module(mod, qconfig, mod_dict, patched_mod=None):
    """Replaces the module with patched module according to mod_dict.

    Args:
        mod (nn.module): The module that will be replaced with a patched module that quantize the inputs/outputs.
        qconfig (ModuleExtraConfig): The quantization config object with the information how to quantize the inputs/outputs.
        mod_dict (dict): dictionary from module name to its patched module.

    Returns:
        nn.module: The new patched module after patching.
    """
    parent = parent_child_mod_dict[mod].parent
    name = parent_child_mod_dict[mod].name
    if patched_mod is None:
        patched_mod = mod_dict[mod.__class__.__name__].patched_module(mod, parent, qconfig)
    setattr(parent, name, patched_mod)


def apply_hf_hook(module):
    """Applies hf_hook on a given module so its weights will be loaded from disk to cpu and then we can quantize it."""
    if hasattr(module, "_hf_hook"):
        module._hf_hook.pre_forward(module)
        module._hf_hook.detach_hook(module)
        delattr(module, "_hf_hook")
    if hasattr(module, "_old_forward"):
        module.forward = module._old_forward
        delattr(module, "_old_forward")


def quantize_params(mod, mod_extra_config):
    """Quantizes the weights of the given module according to the quantization info from mod_extra_config.

    Args:
        mod (nn.module): The module that its weights will be quantized.
        mod_extra_config (ModuleExtraConfig): The quantization config object with the information how to quantize the inputs/outputs.
    """
    for param_name in mod_extra_config.params:
        quantizer = mod_extra_config.params[param_name][0]
        param = getattr(mod, param_name)
        if param.dtype == torch.float16:
            param = param.to(torch.bfloat16)
        param = dequant_original_fp8_weight_if_needed(mod, param)
        quantized_param = quantizer(param.to(cur_accelerator.name()))
        delattr(mod, param_name)
        setattr(mod, param_name, nn.Parameter(quantized_param))
        # Note: in case of re-quantize the fp8 weights, we need to set `updated_fp8_weight` to True
        mod.updated_fp8_weight = True
        quantized_param = getattr(mod, param_name)
        quantized_param.requires_grad_(False)
        cur_accelerator.synchronize()


def convert_fp16_to_bf16(model):
    """Convert all float16 parameters and buffers in the model to bfloat16 after FP8 quantization.
    
    Args:
        model (torch.nn.Module): The PyTorch model that needs to be converted.
    """
    # convert parameters
    for name, param in model.named_parameters():
        if param.dtype == torch.float16:
            param.data = param.data.to(torch.bfloat16)
            logger.debug("Convert FP16 to BF16, parameter name: %s", name)
    
    # convert buffers
    for name, buffer in model.named_buffers():
        if buffer.dtype == torch.float16:
            buffer.data = buffer.data.to(torch.bfloat16)
            logger.debug("Convert FP16 to BF16, buffer name: %s", name)


def prepare_model(model, mod_list, measurement, scale_file, scale_method_config, scale_config):
    """Calculates scales according to the scaling method and config.
    Replaces the model submodules according to the mod_list with patched quantization modules.
    Configures patched modules with the quantization/dequantization methods to apply on their input and output tensors.
    Quantizes the model parameters as they are static.

    Args:
        model (nn.module): The model to quantize.
        mod_list (list): The specific submodules that will be quantized in the model.
        measurement (dict): The measurements of the model.
        scale_file (str): The file containing the scales.
        scale_method_config (dict): The scaling method to use.
        scale_config (dict): The scaling configuration.
    """
    config = get_hqt_config(model)
    recalc_scales = config.cfg["recalc_scales"]
    scales_file_format = np.ndarray
    scales_obj = (
        load_scales(scale_file + ".npz", scales_file_format)
        if (scale_file is not None) and not recalc_scales
        else {}
    )
    scales = convert_scales_to_tensors_dict(scales_obj, scales_file_format, scale_config["hp_dtype"])
    save_file = False
    patched_modules = []
    patched_module_types = set()
    scale_method_config_by_mod_map = {}
    is_dynamic_quantization = config.cfg["dynamic_quantization"]

    should_quantize_cond = True # In static quantization we quantize everything
    with torch.no_grad():
        for name, mod in model.named_modules():
            mod_type_str = mod.__class__.__name__

            if name in mod_list and name not in scales and config.cfg["use_stats_files"] and name not in measurement:
                if mod_default_dict[mod_type_str].should_measure_and_quant:
                    if not config.cfg["ignore_modules_wo_measures"]:
                        patch_module(mod, None, None, PatchedUnmeasuredModule(name, mod))
                    else:
                        logger.debug("Module %s was not quantized.", name)
                    continue
            # When offloading weight to disk, need to transfer the weight from disk to cpu using hf_hook
            apply_hf_hook(mod)
            if name in mod_list:
                set_hqt_config(mod, config)  # set config in the module, as it consumed by the patched module
                if is_dynamic_quantization:
                    # TODO [SW-217813]: support dynamic quantization in all ops and remove supports_dynamic_quant, then move outside the loop
                    should_quantize_cond = is_supported_dynamic_op(mod_type_str)

                # TODO [SW-217813]: support dynamic quantization in all ops and remove should_quantize_cond
                if should_quantize_cond:
                    scale_method_config_node = find_node_scale_method_config(scale_method_config, name, mod_type_str)
                    mod_extra_config, save_file = load_layer_scales(mod, name, config,
                                                                mod_type_str, measurement,
                                                                scales, scale_file,
                                                                scales_file_format,
                                                                scales_obj, scale_method_config_node,
                                                                scale_config, save_file, scale_method_config_by_mod_map)

                    if not config.cfg["fake_quant"] and mod_default_dict[mod_type_str].should_measure_and_quant:
                        quantize_params(mod, mod_extra_config)
                    patch_module(mod, mod_extra_config, mod_default_dict)
                    patched_modules.append(name)
                    patched_module_types.add(type(mod))
                    logger.debug("Patched module name: %s", name)
    if save_file: # cache calculated scales
        save_scales(model, scales_obj, scales_file_format, scale_file + ".npz")
        save_scales(model, scales_obj, scales_file_format, scale_file + ".json")
    scale_method_config_dump_path = os.environ.get("SCALE_METHOD_CONFIG_DUMP_PATH", None)
    if scale_method_config_dump_path is not None:
        if not os.path.exists(os.path.dirname(scale_method_config_dump_path)):
            raise FileNotFoundError(f"Scale method config dump path {scale_method_config_dump_path} does not exist.")
        dump_scale_method_config_by_mod_map(scale_method_config_by_mod_map, scale_method_config_dump_path)
    logger.debug("Patched module types: %s", patched_module_types)
    logger.debug("Patched modules: %s", patched_modules)
    logger.debug("Total patched modules: %d", len(patched_modules))
    model = model.to(cur_accelerator.name())
    convert_fp16_to_bf16(model)
    cur_accelerator.synchronize()


def prepare_model_with_dummy_measurement(model, mod_list, scale_method_config, scale_config):
    """Aim for loading, replace module with patched module for model on meta device.

    Args:
        model (torch.nn.Module): empty model on meta device
        mod_list (list): The specific submodules that will be quantized in the model.
        scale_method_config (dict): The scaling method to use.
        scale_config (dict): The scaling configuration.

    Returns:
        model: empty model that quantized by default qconfig.
    """
    from .patching_common import mod_types
    from ..model_configs import ModuleExtraConfig

    config = get_hqt_config(model)
    patched_modules = []
    patched_module_types = set()
    with torch.no_grad():
        for name, mod in model.named_modules():
            if name not in mod_list:
                continue
            set_hqt_config(mod, config)  # set config in the module, as it consumed by the patched module
            mod_type_str = mod.__class__.__name__
            mode_type = config.cfg["mod_dict"][mod_type_str]
            mod_info = mod_types[mode_type]
            op_obj = ops_quantizer.get_op_quantizer(scale_method_config, mod, None, scale_config, mod_type_str)
            dummy_mod_scales = op_obj.get_scales_module_config()
            dummy_mod_config = op_obj.scales_module_config_to_q_and_dq(dummy_mod_scales)
            dummy_mod_extra_config = ModuleExtraConfig(
                dummy_mod_config.inputs,
                dummy_mod_config.outputs,
                dummy_mod_config.params,
                dummy_mod_scales,
                scale_config,
                )
            # replace bf16 meta weights with FP8 meta weights for loading
            if not config.cfg["fake_quant"] and mod_default_dict[mod_type_str].should_measure_and_quant:
                for param_name in mod_info.param_names:
                    if param_name == "weight":  # only weight is quantized now
                        raw_param = getattr(mod, param_name)
                        param = torch.ones(raw_param.shape, dtype=scale_config["lp_dtype"], device="meta")  # meta tensor
                        delattr(mod, param_name)
                        setattr(mod, param_name, nn.Parameter(param))
            patch_module(mod, dummy_mod_extra_config, mod_default_dict)
            patched_modules.append(name)
            patched_module_types.add(type(mod))
            logger.debug("Patched module name: %s", name)
    logger.debug("Patched module types: %s", patched_module_types)
    logger.debug("Patched modules: %s", patched_modules)
    logger.debug("Total patched modules: %d", len(patched_modules))
    return model


def quantize(model, mod_list):
    """Builds quantization config object that contains for each submodule its quantization functions as preparation for quantization.

    Args:
        model (nn.module): The model that will be quantized.
        mod_list (list, optional): The specific modules that will be quantized in the model.
    """
    config = get_hqt_config(model)
    generate_model_info(model)
    hp_dtype = config.cfg["hp_dtype"]
    lp_dtype = config.cfg["fp8_config"]
    scale_method_config = config.cfg["scale_method"]
    scale_config = config.cfg["scale_params"]
    scale_config["hp_dtype"] = hp_dtype
    scale_config["lp_dtype"] = lp_dtype
    if config.cfg["mode"] == QuantMode.QUANTIZE:
        measurement = {}
        scale_file = None
        use_stats_files = config.cfg["use_stats_files"]
        if use_stats_files:
            measurement = load_measurements(model, config.cfg["measure_file"])
            scale_file = config.cfg["scale_file"]
        prepare_model(model, mod_list, measurement, scale_file, scale_method_config, scale_config)
    elif config.cfg["mode"] == QuantMode.LOAD:
        # no measurement and scale file
        scale_method_config = {CfgStr.ACTIVATION: ScaleMethodConfig(scale_value_type=ScaleValueType.DUMMY_SCALES),
                               CfgStr.WEIGHT: ScaleMethodConfig(scale_value_type=ScaleValueType.DUMMY_SCALES)}
        prepare_model_with_dummy_measurement(model, mod_list, scale_method_config, scale_config)
    else:
        raise Exception("unexpected mode, expected QuantMode.QUANTIZE or QuantMode.LOAD")


# Copyright (c) 2023-2024 Mobiusml and Intel Corporation
#
# This code is based on Mobiusml's HQQ library.
# https://github.com/mobiusml/hqq
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Optimization logic of HQQ."""

import numpy as np
import torch

from neural_compressor.torch.utils import logger
from neural_compressor.torch.utils.auto_accelerator import auto_detect_accelerator


# Proximal solver || W - dequantize(quantize(W))||_p^p
@torch.inference_mode()
def optimize_weights_proximal_legacy(
    tensor,
    scale,
    zero,
    min_max,
    axis=0,
    device="cuda",
    opt_params={"lp_norm": 0.7, "beta": 1e1, "kappa": 1.01, "iters": 20},
    verbose=False,
):
    """Quantize the scale/zero of quantized tensor using the HQQ.

    Args:
        tensor (torch.Tensor): The input tensor to optimize.
        scale (torch.Tensor): The scaling factor for quantization.
        zero (torch.Tensor): The zero-point for quantization.
        min_max (tuple): The minimum and maximum values for quantization.
        axis (int, optional): The axis along which to compute the mean for zero-point calculation. Defaults to 0.
        device (str, optional): The device to use for computation. Defaults to "cuda".
        opt_params (dict, optional): Optimization parameters.
            Defaults to {"lp_norm": 0.7, "beta": 1e1, "kappa": 1.01, "iters": 20}.
        verbose (bool, optional): Whether to print verbose output. Defaults to False.

    Returns:
        tuple: A tuple containing the optimized scale and zero-point tensors.
    """
    lp_norm, beta, kappa, iters = (
        opt_params["lp_norm"],
        opt_params["beta"],
        opt_params["kappa"],
        opt_params["iters"],
    )
    device = auto_detect_accelerator().current_device()

    # TODO: refine it for cpu device
    if auto_detect_accelerator().name() == "cuda":
        dtype = torch.float16
    else:
        dtype = torch.float32
    W_f = tensor.to(dtype).to(device)
    scale = scale.to(dtype).to(device)
    zero = zero.to(dtype).to(device)

    if lp_norm == 1:
        shrink_op = lambda x, beta: torch.sign(x) * torch.nn.functional.relu(torch.abs(x) - 1.0 / beta)
    else:
        shrink_op = lambda x, beta, p=lp_norm: torch.sign(x) * torch.nn.functional.relu(
            torch.abs(x) - (1.0 / beta) * torch.pow(torch.abs(x), p - 1)
        )

    best_error = 1e4
    for i in range(iters):
        W_q = torch.round(W_f * scale + zero).clamp(min_max[0], min_max[1])
        W_r = (W_q - zero) / scale
        W_e = shrink_op(W_f - W_r, beta)
        zero = torch.mean(W_q - (W_f - W_e) * scale, axis=axis, keepdim=True)
        beta *= kappa

        current_error = float(torch.abs(W_f - W_r).mean())
        if verbose:
            logger.info(i, np.round(current_error, 6))
        if current_error < best_error:
            best_error = current_error
        else:
            break

    scale = scale.to(tensor.device)
    zero = zero.to(tensor.device)
    del W_f, W_q, W_r, W_e
    auto_detect_accelerator().empty_cache()

    return scale, zero


optimize_weights_proximal = optimize_weights_proximal_legacy



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""AWQ quantization."""
# Copied from neural_compressor/adaptor/torch_utils/awq.py

import copy
from collections import OrderedDict

import torch

from neural_compressor.torch.algorithms import Quantizer
from neural_compressor.torch.utils import get_accelerator, logger

from .modules import MulLinear
from .utility import (
    fetch_module,
    get_absorb_layers,
    get_block_prefix,
    get_module_input_output,
    recover_forward,
    replace_forward,
    set_module,
)

__all__ = ["AWQQuantizer"]


def _get_absorb_per_block(model, example_inputs, folding=False, weight_config={}):
    """Get absorbed layer per block.

    Args:
        model (torch.nn.Module): input model.
        example_inputs (tensor/tuple/dict, optional): used to trace torch model.
        folding (bool, optional): whether only allow update scale when it can be fold
                                    to upper layer. Defaults to False.
        weight_config (dict, optional): the quantization configuration. Defaults to {}.

    Returns:
        block_absorb_dict: The dict of absorbed layer per block. eg. {0, [[absorbed_1, xx], [xx]], ...}
        absorb_layer_dict: The layer dict that scale can be absorbed. The dict is the inverse of
                                block_absorb_dict for all blocks.
    """
    block_absorb_dict = {}  # record absorbed layer per block
    absorb_layer_dict = {}  # record absorb layers for absorbed layers
    absorb_to_layer, no_absorb_layers = get_absorb_layers(
        model, example_inputs, supported_layers=["Linear"], folding=False
    )
    logger.debug(f"The no absorb layers: {no_absorb_layers}")
    # skip ops when algorithm is not AWQ
    skip_op_set = set()
    for k, v in absorb_to_layer.items():
        for vv in v:
            if vv in weight_config and weight_config[vv]["dtype"] == "fp32":
                skip_op_set.add(k)
    for k in no_absorb_layers:
        if k in weight_config and weight_config[k]["dtype"] == "fp32":
            skip_op_set.add(k)
    for k in skip_op_set:
        if k in absorb_to_layer:
            absorb_to_layer.pop(k)
        if k in no_absorb_layers:
            no_absorb_layers.remove(k)
    if len(skip_op_set) > 0:
        logger.info(f"{skip_op_set} are skipped when running AWQ optimization")

    block_prefix, block_num = get_block_prefix(model)
    for i in range(block_num):
        block_absorb_dict[i] = []
        block_name = block_prefix + "." + str(i) + "."
        for k, v in absorb_to_layer.items():
            name_list = tuple(vv for vv in v if block_name in vv)
            if len(name_list) > 0:
                block_absorb_dict[i].append(name_list)
                absorb_layer_dict[name_list] = k
        if not folding:
            for k in no_absorb_layers:
                if block_name in k:
                    name_list = tuple([k])
                    block_absorb_dict[i].append(name_list)
                    absorb_layer_dict[name_list] = k
    logger.debug(f"The absorbed layers per block: {block_absorb_dict}")
    logger.debug(f"The absorb_layer_dict: {absorb_layer_dict}")
    return block_absorb_dict, absorb_layer_dict


def _get_absorb_dict(model, absorb_layer_dict):
    """Get absorbed layer per block from absorbed layer dict.

    Args:
        model (torch.nn.Module): input model
        absorb_layer_dict (dict): The layer type dict that scale can be absorbed, default is {}.

    Returns:
        block_absorb_dict: dict of absorbed layer per block. eg. {0, [[absorbed_1, xx], [xx]], ...}
        new_absorb_layer_dict: The layer dict that scale can be absorbed. The dict is the inverse of
                                block_absorb_dict for all blocks.
    """
    block_absorb_dict = {}
    block_prefix, block_num = get_block_prefix(model)
    new_absorb_layer_dict = {}
    for i in range(block_num):
        block_absorb_dict[i] = []
        block_name = block_prefix + "." + str(i) + "."

        for k, v in absorb_layer_dict.items():

            if isinstance(v, str):
                name_list = (block_name + v,)
            else:
                name_list = tuple(block_name + vv for vv in v)
            block_absorb_dict[i].append(name_list)
            new_absorb_layer_dict[name_list] = block_name + k
    logger.debug(f"The absorbed layers per block: {block_absorb_dict}")
    logger.debug(f"The absorb_layer_dict: {absorb_layer_dict}")
    return block_absorb_dict, new_absorb_layer_dict


@torch.no_grad()
def _get_weight_scale(weight, q_group_size=-1):
    """Get scale for weight.

    Args:
        weight (tensor): input weight
        q_group_size (int, optional): how many elements share one scale/zp. Defaults to -1.

    Returns:
        scale: the scale of input weight.
    """
    org_shape = weight.shape
    if q_group_size > 0:
        weight = weight.view(-1, q_group_size)
    scale = weight.abs() / weight.abs().amax(dim=1, keepdim=True)
    scale = scale.view(org_shape)
    scale = scale.mean(0)
    return scale


@torch.no_grad()
def _get_act_scale(input_val):
    tmp = [x.abs().view(-1, x.shape[-1]) for x in input_val]
    tmp = torch.cat(tmp, dim=0)
    return tmp.mean(0)


class ActAwareWeightQuant:
    """Implementation of Activation-aware Weight quantization (AWQ) algo."""

    def __init__(
        self,
        model,
        example_inputs=None,
        data_type="int",
        bits=4,
        group_size=32,
        scheme="asym",
        use_full_range=False,
        weight_config={},
        total_block_args=[],
        total_block_kwargs=[],
        device="auto",
        absorb_layer_dict={},
    ):

        self.example_inputs = example_inputs
        self.model = model
        self.device = device
        self._move_model_and_data_to_device()
        self.total_block_args = total_block_args
        self.total_block_kwargs = total_block_kwargs
        # get block list and block prefix, number
        self.block_prefix, self.block_num = get_block_prefix(model)
        self.block_list = fetch_module(model, self.block_prefix)
        self.data_type = data_type
        self.bits = bits
        self.group_size = group_size
        self.scheme = scheme
        self.use_full_range = use_full_range
        self.weight_config = weight_config
        self.absorb_layer_dict = absorb_layer_dict

    def _move_model_and_data_to_device(self):
        # Put the model and example_inputs into target device
        device = get_accelerator(self.device).current_device_name()
        self.model.to(device)
        self.example_inputs = self.example_inputs.to(device)

    def quantize(self, use_auto_scale=True, use_mse_search=True, folding=False, return_int=False):
        """Execute AWQ quantization.

        Args:
            use_auto_scale (bool, optional): whether search scale. Defaults to True.
            use_mse_search (bool, optional): whether search clip range. Defaults to True.
            folding (bool, optional): whether only allow update scale when it can be fold
                                      to upper layer. Defaults to False.
            return_int (bool, optional): whether return int dtype with INCWeightOnlyLinear.
                                         Defaults to False.

        Returns:
            model: quantized model
        """
        # Step 1: get absorbed module list per block, includes self-absorption
        # block_absorb_dict is split per block, includes all absorb relationship.
        # absorb_layer_dict is the inverse of block_absorb_dict for all blocks
        if not self.absorb_layer_dict:
            self.block_absorb_dict, self.absorb_layer_dict = _get_absorb_per_block(
                self.model,
                self.example_inputs,
                # for only use_mse_search, folding is useless.
                folding=folding if use_auto_scale else False,
                weight_config=self.weight_config,
            )
        else:
            self.block_absorb_dict, self.absorb_layer_dict = _get_absorb_dict(self.model, self.absorb_layer_dict)
        # process per block
        for i, module_list in self.block_absorb_dict.items():
            logger.info(f"Processing block: {i+1}/{self.block_num}")
            if len(module_list) == 0:
                logger.info("No need to process this block.")
                continue
            # Step 1: fetch all input values of each linear for scale calculation
            # use the first linear for QKV tuple
            block_name = self.block_prefix + "." + str(i)
            block = fetch_module(self.model, block_name)
            module_hook_config = {v[0].split(block_name + ".")[1]: ["input"] for v in module_list}

            def block_calibration(model):
                for args, kwargs in zip(self.total_block_args, self.total_block_kwargs):
                    model(*args, **kwargs)

            input_values = get_module_input_output(
                block,
                module_hook_config,
                calib_func=block_calibration,
            )
            # Step 3: search best scale for linears in one block and apply it
            if use_auto_scale:
                scale_info = self.search_scale(block, block_name, module_list, input_values)
            # Step 2: update self.total_block_args, self.total_block_kwargs for next block
            out_list = self.block_inference(block)
            self.update_block_input(out_list)
            # Step 4: get input of next block before update scale
            # weights of linear is updated by scale
            if use_auto_scale:
                self.apply_scale(scale_info)
            # Step 5: search best clip range for linears in one block and save to weight_config
            if use_mse_search:
                self.search_clip(block_name, module_list, input_values)
        # Step 6: apply clip range in weight_config when quantizing model weights
        self.apply_quantize_with_clip(return_int)
        return self.model

    def search_scale(self, block, block_name, module_list, input_values):
        """Search scales per block.

        Args:
            block (torch.nn.Module): a block of model
            block_name (str): the block name in model.
            module_list (dict): contains all linear tuple in current block,
                                linears in the same tuple shares scale.
            input_values (dict): contains all input values of linears in current block

        Returns:
            scale_info: a dict that contains input scales of linears in current block
        """
        from .utility import quant_tensor

        scale_info = {}
        logger.info("Searching best scales with AWQ algorithm")
        for module_tuple in module_list:
            # Step 1: Initialize quantization configuration.
            if module_tuple[0] in self.weight_config:
                cur_dtype = self.weight_config[module_tuple[0]]["dtype"]
                cur_bits = self.weight_config[module_tuple[0]]["bits"]
                cur_group_size = self.weight_config[module_tuple[0]]["group_size"]
                cur_scheme = self.weight_config[module_tuple[0]]["scheme"]
            else:
                cur_dtype, cur_bits, cur_group_size, cur_scheme = (
                    self.data_type,
                    self.bits,
                    self.group_size,
                    self.scheme,
                )
            if cur_bits < 0:
                continue
            logger.info(f"[SCALE] Processing module: {module_tuple}")
            # Step 2: update module name in block
            module_name_list = [i.split(block_name + ".")[1] for i in module_tuple]
            # Step 3: collect w_max and x_max for scale calculation.
            weight = torch.cat([fetch_module(block, _m).weight for _m in module_name_list], dim=0)
            w_max = _get_weight_scale(weight, q_group_size=cur_group_size)
            del weight
            input_val = input_values[module_name_list[0]]["input"]
            x_max = _get_act_scale(input_val)
            absorbed_modules = {_m: fetch_module(block, _m) for _m in module_name_list}
            # Step 4: collect origin output for MSE and state_dict for recover.
            org_stat = {_m: copy.deepcopy(module.state_dict()) for _m, module in absorbed_modules.items()}
            if len(module_tuple) > 1:
                # use block inference for multi-modules
                org_out = self.block_inference(block)
            else:
                module = absorbed_modules[module_name_list[0]]
                org_out = self.module_inference(module, input_val)
            # Step 5: collect origin output for MSE and state_dict for recover.
            best_error = float("inf")
            best_scales = None
            best_scale_alpha = None
            n_grid = 20
            history = []
            # Step 6: set different alpha for scale and compare the MSE loss.
            for ratio in range(n_grid):
                ratio = ratio * 1 / n_grid
                scales = (x_max.pow(ratio) / w_max.pow(1 - ratio)).clamp(min=1e-4).view(-1)
                scales = scales / (scales.max() * scales.min()).sqrt()
                for name, module in absorbed_modules.items():
                    module.weight.data = module.weight.data.mul(scales.view(1, -1))
                    module.weight.data = quant_tensor(
                        module.weight.data,
                        data_type=cur_dtype,
                        num_bits=cur_bits,
                        group_size=cur_group_size,
                        scheme=cur_scheme,
                        full_range=self.use_full_range,
                    ) / scales.view(1, -1)
                loss = 0
                if len(module_tuple) > 1:
                    # use block inference for multi-modules
                    cur_out = self.block_inference(block)
                else:
                    module = absorbed_modules[module_name_list[0]]
                    cur_out = self.module_inference(module, input_val)
                for out1, out2 in zip(org_out, cur_out):
                    loss += (out1 - out2).float().pow(2).mean().item()
                history.append(loss)
                is_best = loss < best_error
                if is_best:
                    best_error = loss
                    best_scales = scales
                    best_scale_alpha = ratio
                for name, module in absorbed_modules.items():
                    module.load_state_dict(org_stat[name])
            # Step 7: record the best scale alpha of each module_tuple
            assert best_scales is not None, "Loss is infinity! Cannot find the correct scale."
            best_scales = best_scales.view(-1)
            assert torch.isnan(best_scales).sum() == 0, best_scales
            scales = best_scales.detach()
            scale_info[module_tuple] = scales
            logger.debug("The loss history of different scale:{}".format(history))
            logger.info("The best scale alpha of {}: {}".format(module_tuple, best_scale_alpha))
        return scale_info

    @torch.no_grad()
    def apply_scale(self, scale_info):
        """Apply scales to model.

        Args:
            scale_info (dict): a dict that contains input scales of linears in current block
        """
        for module_tuple, scale in scale_info.items():
            logger.debug(f"apply scale for module: {module_tuple}")
            assert module_tuple in self.absorb_layer_dict, "cannot find the absorb module."
            absorb_module_name = self.absorb_layer_dict[module_tuple]
            absorb_module = fetch_module(self.model, absorb_module_name)
            if absorb_module_name == module_tuple[0]:
                # Case 1: module is self-absorption
                new_module = MulLinear(absorb_module, 1.0 / scale)
                new_module._update_linear()
                set_module(self.model, absorb_module_name, new_module)
            else:
                # Case 2: scale is absorbed by other layer
                if len(absorb_module.weight.shape) == 1:
                    absorb_module.weight.div_(scale)  # for LayerNorm
                else:
                    absorb_module.weight.div_(scale.view(-1, 1))
                # hasattr is for LlamaRMSNorm
                if hasattr(absorb_module, "bias") and absorb_module.bias is not None:
                    absorb_module.bias.div_(scale.view(-1))
                for name in module_tuple:
                    absorbed_module = fetch_module(self.model, name)
                    absorbed_module.weight.mul_(scale.view(1, -1))

    def search_clip(self, block_name, module_list, input_values):
        """Search best clip range of each linears in current block.

        Args:
            block_name (str): block name in model.
            module_list (dict): contains all linear tuple in current block,
                                linears in the same tuple shares scale.
            input_values (dict): contains all input values of linears in current block
        """
        from .utility import quant_tensor

        logger.info("Searching the best clip range with AWQ algorithm")
        for module_tuple in module_list:
            input_val = input_values[module_tuple[0].split(block_name + ".")[1]]["input"]
            # process linear modules one by one
            for module_name in module_tuple:
                # Step 1: Initialize quantization configuration.
                if module_name in self.weight_config:
                    cur_dtype = self.weight_config[module_name]["dtype"]
                    cur_bits = self.weight_config[module_name]["bits"]
                    cur_group_size = self.weight_config[module_name]["group_size"]
                    cur_scheme = self.weight_config[module_name]["scheme"]
                else:
                    cur_dtype, cur_bits, cur_group_size, cur_scheme = (
                        self.data_type,
                        self.bits,
                        self.group_size,
                        self.scheme,
                    )
                if cur_bits < 0:
                    continue
                logger.info(f"[CLIP] Processing module: {module_name}")
                # Step 2: update module name
                module = fetch_module(self.model, module_name)
                # Step 3: collect origin output for MSE and state_dict for recover.
                org_stat = copy.deepcopy(module.state_dict())
                org_out = self.module_inference(module, input_val)
                # Step 4:  set different clip range for weight and compare the MSE loss.
                logger.info("Searching the best clip range with AWQ algorithm")
                best_error = float("inf")
                best_clip_ratio = None
                n_grid = 100
                max_shrink = 0.1
                history = []
                for i_s in range(int(max_shrink * n_grid)):
                    ratio = 1 - i_s / n_grid  # 1, 0.91-1.0
                    # MulLinear can also work with @weight.setter
                    module.weight.data = quant_tensor(
                        module.weight.data,
                        data_type=cur_dtype,
                        num_bits=cur_bits,
                        group_size=cur_group_size,
                        scheme=cur_scheme,
                        full_range=self.use_full_range,
                        quantile=ratio,
                    )
                    loss = 0
                    cur_out = self.module_inference(module, input_val)
                    for out1, out2 in zip(org_out, cur_out):
                        loss += (out1 - out2).float().pow(2).mean().item()
                    history.append(loss)
                    is_best = loss < best_error
                    if is_best:
                        best_error = loss
                        best_clip_ratio = ratio
                    module.load_state_dict(org_stat)
                logger.debug("The loss history of different clip range:{}".format(history))
                if module_name not in self.weight_config:
                    self.weight_config[module_name] = {
                        "bits": cur_bits,
                        "group_size": cur_group_size,
                        "scheme": cur_scheme,
                    }
                self.weight_config[module_name]["quantile"] = best_clip_ratio
                if isinstance(module, MulLinear):
                    self.weight_config[module_name + ".linear"] = self.weight_config[module_name]
                    self.weight_config.pop(module_name)
                logger.debug("The best clip ratio for {}:{}".format(module_name, best_clip_ratio))

    def apply_quantize_with_clip(self, return_int=False):
        """Quantize model with clip range.

        Args:
            return_int (bool, optional): whether return int dtype with INCWeightOnlyLinear.
                                         Defaults to False.
        """
        # apply quantization and clip
        logger.info("Quantizing the AWQ optimized fp32 model")
        from .rtn import RTNQuantizer

        rtn_quantizer = RTNQuantizer(quant_config=self.weight_config)

        self.model = rtn_quantizer.quantize(
            self.model,
            bits=self.bits,
            group_size=self.group_size,
            scheme=self.scheme,
            return_int=return_int,
            use_full_range=self.use_full_range,
        )
        logger.info("AWQ quantization is done.")

    def update_block_input(self, input_list):
        """Update block input for next block inference.

        Args:
            input_list (list): A list of previous block outputs to serve as input to the next block.
        """
        for i, inp in enumerate(input_list):
            if len(self.total_block_args[i]) > 0:
                self.total_block_args[i][0] = inp
            elif "hidden_states" in self.total_block_kwargs[i]:
                self.total_block_kwargs[i]["hidden_states"] = inp
            else:  # pragma: no cover
                assert False, "cannot find hidden_states position for next block"

    def block_inference(self, model):
        """Collect output of block.

        Args:
            model (torch.nn.Module): input model.

        Returns:
            output(list):  a list of block output.
        """
        total_out = []
        for args, kwargs in zip(self.total_block_args, self.total_block_kwargs):
            # to avoid layer_past: Dynamic_cache when transformers higher than 4.45.1
            if "layer_past" in kwargs.keys() and kwargs["layer_past"] is not None:
                kwargs["layer_past"] = None
            out = model(*args, **kwargs)
            if isinstance(out, tuple):  # pragma: no cover
                out = out[0]
            total_out.append(out)
        return total_out

    def module_inference(self, model, inputs):
        """Collect output of module.

        Args:
            model (torch.nn.Module): input model.
            inputs (list): a list of module input.

        Returns:
            output(list):  a list of module output.
        """
        total_out = []
        for inp in inputs:
            out = model(inp)
            if isinstance(out, tuple):  # pragma: no cover
                out = out[0]
            total_out.append(out)
        return total_out


class AWQQuantizer(Quantizer):
    """AWQ Quantizer."""

    def __init__(self, quant_config: OrderedDict = {}, absorb_layer_dict: dict = {}):
        """Init an AWQQuantizer object.

        Args:
            quant_config (OrderedDict, optional): quantization config for ops. Defaults to {}.
            absorb_layer_dict (dict): The layer dict that scale can be absorbed, default is {}.
        """
        super().__init__(quant_config)
        self.absorb_layer_dict = absorb_layer_dict

    @torch.no_grad()
    def prepare(self, model, *args, **kwargs):
        """Prepare a given model to get hidden states and kwargs of first block.

        Args:
            model: A float torch model.

        Returns:
            A prepared model.
        """
        assert isinstance(model, torch.nn.Module), "AWQ algorithm only supports torch module"
        model = replace_forward(model)
        return model

    @torch.no_grad()
    def convert(
        self,
        model,
        bits=4,
        group_size=32,
        scheme="asym",
        example_inputs=None,
        use_auto_scale=True,
        use_mse_search=True,
        folding=False,
        return_int=False,
        use_full_range=False,
        data_type="int",
        *args,
        **kwargs,
    ):
        """Converts a prepared model to a quantized model.

        Args:
            model: torch model.
            bits: num bits. Defaults to 4.
            group_size: how many elements share one scale/zp. Defaults to 32.
            scheme: sym or asym. Defaults to "asym".
            example_inputs: example_inputs. Defaults to None.
            use_auto_scale: whether enable scale for salient weight. Defaults to True.
            use_mse_search: whether enable clip for weight by checking mse. Defaults to True.
            folding: False will allow insert mul before linear when the scale cannot be absorbed
                by last layer, else won't. Defaults to False.
            return_int: Choose return fp32 or int32 model. Defaults to False.
            use_full_range: Choose sym range whether use -2**(bits-1). Defaults to False.
            data_type: data type. Defaults to "int".

        Returns:
            model: fake quantized model
        """
        model = recover_forward(model)
        total_block_args = getattr(model, "total_block_args", [])
        total_block_kwargs = getattr(model, "total_block_kwargs", [])
        delattr(model, "total_block_args")
        delattr(model, "total_block_kwargs")

        awq = ActAwareWeightQuant(
            model,
            example_inputs=example_inputs,
            data_type=data_type,
            bits=bits,
            group_size=group_size,
            scheme=scheme,
            use_full_range=use_full_range,
            weight_config=self.quant_config,
            total_block_args=total_block_args,
            total_block_kwargs=total_block_kwargs,
            absorb_layer_dict=self.absorb_layer_dict,
        )
        qdq_model = awq.quantize(
            use_auto_scale=use_auto_scale,
            use_mse_search=use_mse_search,
            folding=folding,
            return_int=return_int,
        )
        return qdq_model



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base class and helper functions for registering observers."""

from typing import Dict, Optional, Any
from abc import abstractmethod, ABC
import os
import torch
from neural_compressor.torch.utils.auto_accelerator import auto_detect_accelerator
from neural_compressor.torch.algorithms.fp8_quant.model_configs import (
    ModuleConfig,
    OBSERVER_TYPES,
    OBSERVER_PARAMS,
    IMOD_DICT,
)
from neural_compressor.torch.algorithms.fp8_quant.utils.logger import logger
from neural_compressor.torch.algorithms.fp8_quant._quant_common.quant_config import get_hqt_config

__all__ = [
    "ObserverBase",
    "register_observer",
]

cur_device = auto_detect_accelerator().current_device_name()

class ObserverBase(ABC):
    def __init__(self, name, mod, d_shape=None, params=None, device=cur_device):
        self.name = name
        self.mod = mod
        self.state = None
        self.used = False
        self.device = device

    @abstractmethod
    def measure(self, x):
        raise NotImplementedError("`measure` is not implemented")

    def is_used(self):
        return self.used

    @abstractmethod
    def update_state(self, x):
        raise NotImplementedError("`update_state` is not implemented")


def register_observer(observer_type):
    """Decorate and register all observer subclasses.

    Args:
        observer_type (str): The observer registration name.
    """
    def decorator(cls):
        if observer_type in OBSERVER_TYPES:
            logger.info("Overwrite the existing observer {}.".format(observer_type))
        OBSERVER_TYPES[observer_type] = cls
        return cls

    return decorator


def register_module_config_for_observer(module_name, inputs_param=(None,), outputs_param=(None,), weight_param=None, **kwargs):
    """Decorate and register module config for specific observer.

    Args:
        module_name (str): The module registration name.
        inputs_param (tuple): The parameter of inputs. It is a tuple of dicts: (input1 param, input2 param, ...).
        outputs_param (tuple): The parameter of outputs. It is a tuple of dicts: (output1 param, output2 param, ...).
        weight_param (dict): The parameter of weight, can be used by some modules like linear.
    """
    def decorator(cls):
        if cls not in OBSERVER_TYPES.values():
            raise ValueError(
                f"Please register observer first and then register module_config like below: \n"
                f"@register_module_config_for_observer(module_name=module_name, inputs_param=..., outputs_param=..., ...) \n"
                f"@register_observer(observer_type=observer_type) \n"
                f"class CustomizedObserver(ObserverBase): \n"
                f"... \n"
            )
        observer_type = list(OBSERVER_TYPES.keys())[list(OBSERVER_TYPES.values()).index(cls)]
        if observer_type in OBSERVER_PARAMS and module_name in OBSERVER_PARAMS[observer_type]:
            logger.info("Overwrite the existing param of {} for {} observer.".format(module_name, observer_type))
        if weight_param is not None:
            kwargs.update({"weight": weight_param})
        OBSERVER_PARAMS.setdefault(observer_type, {})[module_name] = ModuleConfig(
            inputs=inputs_param, outputs=outputs_param, params=kwargs,
        )
        return cls

    return decorator



@register_observer(observer_type="maxabs")
class MaxAbsObserver(ObserverBase):

    def __init__(self, name, mod, d_shape=None, params=None, device=cur_device):
        super().__init__(name=name, mod=mod, device=device)
        self.first = True
        config = get_hqt_config(mod).cfg
        self.state = torch.zeros((1, 1), device=self.device, dtype=config["hp_dtype"])

    def update_state(self, x):
        self.state.copy_(torch.maximum(torch.max(torch.abs(x)).detach(), self.state))

    def measure(self, x):
        self.update_state(x)
        self.used = True


@register_module_config_for_observer(module_name="linear", inputs_param=({"dim": -1},), outputs_param=({"dim": -1},), weight_param={"dim": 0})
@register_module_config_for_observer(module_name="matmul", inputs_param=({"dim": -1}, {"dim": -2},), outputs_param=({"dim": -1},))
@register_observer(observer_type="maxabs_per_channel")
class MaxAbsPerChannelObserver(ObserverBase):
    def __init__(self, name, mod, d_shape=None, params=None, device=cur_device):
        super().__init__(name=name, mod=mod, device=device)
        self.first = True
        self.dim = params["dim"] if (params is not None) and ("dim" in params) else -1
        if d_shape is not None:
            p = list(range(len(d_shape)))
            self.dim = self.dim if self.dim >= 0 else len(d_shape) + self.dim
            p[-1] = self.dim
            p[self.dim] = len(d_shape) - 1
            self.p = p
            self.state = self.init_state_from_shape(d_shape)

    def init_state(self, x):
        device = x.device
        Nch = x.shape[self.dim]
        self.Nch = Nch
        state = torch.zeros((Nch, 1), device=device, dtype=torch.float32)
        self.shape = list(x.shape)
        return state

    def init_state_from_shape(self, x_shape, device=cur_device):
        device = device
        Nch = x_shape[self.dim]
        self.Nch = Nch
        state = torch.zeros((Nch, 1), device=device, dtype=torch.float32)
        self.first = False
        return state

    def update_state(self, x):
        self.state.copy_(
            torch.maximum(
                torch.max(
                    torch.abs(x.permute(self.p).reshape([-1, self.Nch])),
                    dim=0,
                    keepdim=True,
                )[0].t(),
                self.state,
            )
        )

    def measure(self, x):
        if self.first:
            self.state = self.init_state(x)
            self.first = False
        self.update_state(x)
        self.used = True


@register_observer(observer_type="save")
class SaveObserver(ObserverBase):
    def __init__(self, name, mod, d_shape=None, params=None, device=cur_device):
        super().__init__(name=name, mod=mod, device=device)
        self.first = True
        self.cnt = -1
        config = get_hqt_config(mod).cfg
        self.folder_name = os.path.join(config["dump_stats_base_path"], "tensors")
        os.makedirs(self.folder_name, exist_ok=True)
        self.file_base_name = os.path.join(self.folder_name, IMOD_DICT[mod] + "_" + name + "_iter")
        self.state = self.init_state_from_shape(d_shape)

    def init_state(self, x):
        device = x.device
        state = torch.zeros((1, 1), device=device, dtype=torch.float32)
        self.shape = list(x.shape)
        return state

    def init_state_from_shape(self, x_shape, device=cur_device):
        state = torch.zeros((1, 1), device=device, dtype=torch.float32)
        self.first = False
        return state

    def update_state(self, x):
        self.cnt += 1
        torch.save(x, self.file_base_name + str(self.cnt) + ".pt")

    def measure(self, x):
        self.update_state(x)
        self.used = True


@register_observer(observer_type="shape")
class ShapeObserver(ObserverBase):
    def __init__(self, name, mod, d_shape=None, params=None, device=cur_device):
        super().__init__(name=name, mod=mod, device=device)

    def init_state(self, x):
        device = x.device
        Ndim = len(x.shape)
        self.Ndim = Ndim
        state = torch.tensor(x.shape, device=device, dtype=torch.int32).reshape((1, Ndim))
        return state

    def init_state_from_shape(self, x_shape, device=cur_device):
        logger.info("ShapeObserver doesn't support init_state_from_shape")
        return

    def update_state(self, x):
        logger.info("ShapeObserver doesn't support update_state")
        return

    def measure(self, x):
        self.state = self.init_state(x)

    def is_used(self):
        return self.state is not None